from pathlib import Path
import os

class FileExistsException(Exception):
    pass

class ProjectGenerator:
    def __init__(self, force: bool):
        self.force: bool = force
    
    def mkpackage(self, path: Path):
        try: os.mkdir(path) 
        except: ...
        
        init_path = path / "__init__.py"
        
        self.mkmodule(init_path, "")

    def mkmodule(self, path: Path | str, text: str):
        if not self.force and os.path.exists(path):
            raise FileExistsException(f"Файл существует: {path!s}")
        
        with open(path, mode="w", encoding="utf-8") as fh:
            fh.write(text)


import pathlib
from typing import Self
from telegram import Bot, InputFile
import telegram
from telegram import Message as PTBMessage

from ..button_rows import ButtonRows

from ...callback_data import CallbackDataMapping
from ...message import AudioMessage     as BaseAudioMessage
from ...message import SentAudioMessage     as BaseSentAudioMessage
from .message import HasButtonRows, Message, SentMessage


# TODO: Доработать классы, сейчас они вообще не готовы

class AudioMessage(BaseAudioMessage, HasButtonRows, Message):
    def __init__(self, 
            audio: InputFile | bytes | pathlib.Path | telegram.Audio, 
            caption: str, button_rows: ButtonRows = None, parse_mode: str = None):
        super().__init__(caption, button_rows, parse_mode)
        self.audio = audio
    
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_audio(user_id, self.audio
            , caption = self.caption
            , reply_markup=self.get_reply_markup(mapping)\
            , parse_mode = self.parse_mode)
        return SentAudioMessage( # TODO: fix
            self.text, self.button_rows, ptb_message)
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.audio == other.audio and \
            self.button_rows == other.button_rows
    
    def clone(self):
        button_rows = None
        if self.button_rows:
            button_rows = self.button_rows.clone()
        return AudioMessage(self.audio, self.caption, button_rows, 
            self.parse_mode)

class SentAudioMessage(BaseSentAudioMessage, HasButtonRows, SentMessage):
    def __init__(self, 
            audio: str, 
            button_rows: ButtonRows
        , ptb_message: PTBMessage):
        super().__init__(button_rows)
        self.ptb_message = ptb_message 
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if orig.text == self.text and orig.reply_markup == reply_markup:
            return
        self.ptb_message = await bot.edit_message_text(
            text = self.text,
            reply_markup = reply_markup,
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id)
    
    async def delete(self, bot: Bot):
        await bot.delete_message(
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id)
    
    def __eq__(self, other: Self):
        return self.text == other.text and \
            self.button_rows == other.button_rows
    
    def clone(self):
        return SentAudioMessage(self.text, self.button_rows, self.ptb_message)

    def get_unsent(self):
        return AudioMessage(
              self.text
            , self.button_rows)


from abc import ABC
from typing import Callable

class InputCallback: ...

class FuncCallback(InputCallback):
    def __init__(self, function: Callable
        , one_time: bool = True, **kwargs):
        self.function = function
        self.one_time = one_time
        self.kwargs = kwargs
    
    def __call__(self, **kwds):
        return self.function(**self.kwargs, **kwds)

class ScreenCallback(InputCallback):
    def __init__(self, screen_name: str, stack: bool = False):
        self.screen_name = screen_name
        self.stack = stack


import os
import argparse
from pathlib import Path
from .project_generator.common import FileExistsException, ProjectGenerator
from .project_generator.create_init import create_init
from .project_generator.create_model import create_model
from .project_generator.create_screen import create_screen
from .project_generator.create_types import create_types



def main():
    parser = argparse.ArgumentParser(
        description='python -m tg_bot_screen', 
        usage='%(prog)s [options]')
    parser.add_argument("--ptb", "--python-telegram-bot", action="store_true")
    parser.add_argument("-f", "--force", action="store_true")
    
    args = parser.parse_args()
    
    if args.ptb is False:
        good_use = R"python -m tg_bot_screen --ptb"
        print(f"""
Ошибка: Не передан аргумент, который отражает библиотеку, которая будет использована
Например --ptb
\tИспользование:
\t{good_use}
""")
        return
    
    proj_gen = ProjectGenerator(args.force)
    
    try:
        proj_gen.mkmodule("run.py", "from src.init import main")
        basewd = Path("src") # Working Directory
        
        proj_gen.mkpackage(basewd)
        
        create_init(proj_gen, basewd / "init")
        
        create_model(proj_gen, basewd / "model")
        
        create_screen(proj_gen, basewd / "screen")
        
        create_types(proj_gen, basewd / "types")
    except FileExistsException as e:
        print(f"""
Ошибка: {e!s}
Используйте -f или --force для перезаписи
""")


if __name__ == "__main__":
    main()



from pathlib import Path
from .common import ProjectGenerator

def create_init(proj_gen: ProjectGenerator, cwd: Path): 
    proj_gen.mkpackage(cwd)
    proj_gen.mkmodule(cwd / "app.py", """\
from os import environ
from telegram.ext import Application
from src.model.bot_manager import BotManager

token = environ.get("BOT_TOKEN")

application = Application.builder().token(token).build()

bot = application.bot

botm = BotManager(application).build()""")
    
    
    proj_gen.mkmodule(cwd / "main.py", """\
from telegram.ext import CommandHandler
from .app import botm, application
from .screens import load_screens
from .start import start, start_inner

load_screens()

botm.start_inner = start_inner

# application.job_queue.run_repeating(action, interval=60, first=1)

application.add_handler(CommandHandler("start", start), 0)
botm.add_handlers()

print("Запрашивание...")
application.run_polling(0.1)""")

    proj_gen.mkmodule(cwd / "screens.py", R"""\
import importlib
from pathlib import Path

def load_screens():
    p = Path("src/screen")
    for path in p.rglob("*.py"):
        if path.name.startswith("_"):
            continue
        fullpath = str(path)
        module_name = fullpath[:-3].replace("/", ".").replace("\\", ".")
        importlib.import_module(module_name)""")
    
    proj_gen.mkmodule(cwd / "start.py", """\
from telegram import Update, User
from src.init.app import botm

async def start(update: Update, _):
    chat = update.effective_chat
    if chat and chat.type != "private":
        print(f"{chat.id} написал /start не в личном чате")
        return
    
    if update.message:
        user = update.message.from_user
    elif update.callback_query:
        user = update.callback_query.from_user
    else:
        print("Пользователь написал, но не было ни message, ни callback_query")
        return
    assert isinstance(user, User)
    
    user_id = user.id
    await start_inner(user_id)

async def start_inner(user_id: int):  
    if botm.config.development_mode and \
        user_id not in botm.config.admin_list:
        print(f"{user_id} написал /start и не был допущен [development mode]")
        return
    
    print(f"{user_id} написал /start")
    
    sud = botm.system_user_data.get(user_id)
    if sud and sud.screen:
        try: await sud.screen.delete(botm.bot)
        except: ...
    
    botm.system_user_data.reset(user_id)
    botm.reset_user_data(user_id)
    
    await botm.screen.set_by_name(user_id, "welcome")""")


from .common import *

def create_model(proj_gen: ProjectGenerator, cwd: Path): 
    proj_gen.mkpackage(cwd)
    proj_gen.mkmodule(cwd / "bot_manager.py", """\
from typing import Callable
from telegram.ext import Application
from tg_bot_screen.ptb import BotManager as BaseBotManager
from .user_data import UserDataManager
from .config_manager import ConfigManager

class BotManager(BaseBotManager):
    def __init__(self, application: Application):
        super().__init__(application)
        self.user_data_m = UserDataManager()
        self.config = ConfigManager()
        self.config.load()
        
        self.start_inner: Callable = None
    
    def get_user_data(self, user_id: int):
        return self.user_data_m.get(user_id)

    def reset_user_data(self, user_id: int):
        self.user_data_m.reset(user_id)
        
    async def mapping_key_error(self, user_id: int):
        await self.start_inner(user_id)""")
    
    proj_gen.mkmodule(cwd / "config_manager.py", """\
from json import dumps, loads
from pathlib import Path
from typing import Any

class ConfigManager:
    def __init__(self, path: str = "config/config.json"):
        self.__path = Path(path)
        self.__json: dict[str, Any]
        self.defaults = {
            "development_mode" : False,
            "admin_list": []
        }
    
    def load(self):
        file_path = self.__path
        file_path.parent.mkdir(parents=True, exist_ok=True)
        defaults = self.defaults
        check_fields = False # Для оптимизации
        if file_path.exists():
            try:
                json = loads(file_path.read_text("utf-8"))
                check_fields = True
            except Exception as e:
                print(f"ConfigManager Error {e!r}")
                json = defaults
        else:
            json = defaults
        
        self.__json = json

        if check_fields:
            for name in defaults:
                if name not in json:
                    json[name] = defaults[name]
            self.dump_to_file()

    def dump_to_file(self):
        self.__path.touch(exist_ok=True)
        self.__path.write_text(
            dumps(self.__json, indent=4, ensure_ascii=False), 
            encoding="utf-8")
    
    @property
    def development_mode(self) -> bool:
        return self.__json["development_mode"]
    
    @property
    def admin_list(self) -> list[int]:
        return self.__json["admin_list"]
        
    """)
    proj_gen.mkmodule(cwd / "user_data.py", """\
class UserData:
    def __init__(self):
        self.last_error: str = None 

class UserDataManager:
    def __init__(self):
        self.users_data: dict[int, UserData] = {}
    
    def get(self, user_id: int):
        if user_id not in self.users_data:
            ud = UserData()
            self.users_data[user_id] = ud
            return ud
        
        return self.users_data[user_id]

    def reset(self, user_id: int):
        self.users_data[user_id] = UserData()""")